"""自定义常用 URL 方法"""

from colorama import Fore, Style, init
from tqdm.asyncio import tqdm
import aiohttp
import asyncio
import os
import re
import shutil
import socket
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Union, List, Tuple
import pyspeedtest
import requests

from .color_print import Print
from .get_tool_delta_version import get_tool_delta_version

# 使用方法 mirror_github[value: int].format(url: str)
mirror_github = [
    "https://tdload.tblstudio.cn/{}",
    "https://hub.gitmirror.com/{}",
    "https://gh.con.sh/{}",
    "https://mirror.ghproxy.com/{}",
]

# Initialize colorama
init(autoreset=True)


async def download_file_urls(download_url2dst: List[Tuple[str, str]]) -> None:
    """
    从给定的URL并发下载文件到指定的目标路径。

    Args:
        download_url2dst (List[Tuple[str, str]]): 包含多个元组的列表，每个元组包含：
            - url (str): 要下载的文件的URL。
            - dst (str): 下载的文件将保存的目标路径。

    Returns:
        None
    """

    async def download_file(
        session: aiohttp.ClientSession,
        url: str,
        i: int,
        sem: asyncio.Semaphore,
        sem2: asyncio.Semaphore,
        file_path: str,
    ) -> tqdm:
        """
        从给定的URL下载单个文件到指定的目标路径。

        Args:
            session (aiohttp.ClientSession): 用于发送HTTP请求的aiohttp客户端会话。
            url (str): 要下载的文件的URL。
            i (int): 下载任务的索引，用于进度条定位。
            sem (asyncio.Semaphore): 用于限制同时进行的下载任务数量的信号量。
            sem2 (asyncio.Semaphore): 用于限制显示的进度条数量的信号量。
            file_path (str): 下载的文件将保存的目标路径。

        Returns:
            tqdm: 下载任务的进度条。
        """
        async with sem2:
            progress_bar = tqdm(
                desc=f"• Downloading {Fore.CYAN}{url.split('/')[-1]}{Style.RESET_ALL}: {Fore.YELLOW}Pending...{Style.RESET_ALL}",
                total=0,
                unit="MB",
                unit_scale=True,
                bar_format="{desc} {n:.2f}MB/{total:.2f}MB",
                position=i,
            )
            async with sem:
                async with session.get(url) as response:
                    if response.status == 200:
                        filename = url.split("/")[-1]
                        total_size = int(response.headers.get("content-length", 0))
                        total_size_mb = total_size / (1024 * 1024)  # 转换为 MB
                        progress_bar.reset(total=total_size_mb)

                        progress_bar.set_description_str(
                            f"• Downloading {Fore.CYAN}{filename}{Style.RESET_ALL}: {Fore.YELLOW}In Progress...{Style.RESET_ALL}"
                        )
                        downloaded = 0

                        with open(file_path, "wb") as f:
                            async for chunk in response.content.iter_chunked(1024):
                                f.write(chunk)
                                downloaded += len(chunk)
                                progress_bar.update(
                                    len(chunk) / (1024 * 1024)
                                )  # 更新进度为 MB

                        progress_bar.set_description_str(
                            f"• Downloading {Fore.CYAN}{filename}{Style.RESET_ALL}: {Fore.GREEN}Succeed{Style.RESET_ALL}"
                        )
                        progress_bar.bar_format = "{desc}"  # 只显示描述
                        progress_bar.refresh()
                    else:
                        progress_bar.set_description_str(
                            f"• Downloading {Fore.CYAN}{url.split('/')[-1]}{Style.RESET_ALL}: {Fore.RED}Failed (HTTP {response.status}){Style.RESET_ALL}"
                        )
                        progress_bar.bar_format = "{desc}"  # 只显示描述
                        progress_bar.refresh()
            return progress_bar

    sem: asyncio.Semaphore = asyncio.Semaphore(4)  # 限制同时进行的下载任务数量为4
    sem2: asyncio.Semaphore = asyncio.Semaphore(10)  # 只显示 10 个下载任务

    async with aiohttp.ClientSession() as session:
        tasks: List[asyncio.Task] = []
        progress_bars: List[tqdm] = []

        for i, (url, dst) in enumerate(download_url2dst):
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            task: asyncio.Task = asyncio.create_task(
                download_file(session, url, i, sem, sem2, dst)
            )
            tasks.append(task)

        progress_bars = await asyncio.gather(*tasks)
        for progress_bar in progress_bars:
            progress_bar.close()


def format_mirror_url(url: str) -> list:
    """填充 url 到镜像 url 列表

    Args:
        url (str): 原始 URL

    Returns:
        list: 填充原始 url 后的镜像列表
    """
    mir_url: list = []
    for mirror in mirror_github:
        mir_url.append(mirror.format(url))
    return mir_url


def githubdownloadurl_to_rawurl(url: str) -> str:
    """将 GitHub 下载链接转换为原始链接

    Args:
        url (str): 原始链接

    Returns:
        str: 原始链接
    """
    try:
        if url.startswith("https://github.com/"):
            return requests.head(url, allow_redirects=True).url
        return url
    except Exception:
        return url


def progress_bar(
    current: float | int,
    total: float | int,
    length: int | float = 20,
    color1: str = "§f",
    color2: str = "§b",
) -> str:
    """执行进度条

    Args:
        current (float | int): 当前进度值
        total (float | int): 总进度值
        length (int): 进度条长度.
        color1 (str): 进度条颜色 1.
        color2 (str): 进度条颜色 2.

    Returns:
        str: 格式化后的进度条字符串
    """
    pc = round(current / total * length)
    return Print.colormode_replace(
        color1 + " " * pc + color2 + " " * (20 - pc) + "§r ", 7
    )


def download_progress_bar(
    current_bytes: int, total_bytes: int, speed: float = 0
) -> None:
    """构建下载进度条

    Args:
        current_bytes (int): 当前已下载的字节数
        total_bytes (int): 文件总字节数
        speed ( float): 下载速度.
    """
    progressBar = progress_bar(current_bytes, total_bytes)
    b = f"{progressBar} {pretty_kb(current_bytes)}B / {pretty_kb(total_bytes)}B"
    if speed != 0:
        b += f" ({pretty_kb(speed)}B/s)    "
    with Print.lock:
        Print.print_with_info(b, "§a 下载 ", need_log=False, end="\r")


def pretty_kb(n: float) -> str:
    """将字节数转换为可读性更好的字符串表示形式

    Args:
        n (float): 字节数

    Returns:
        str: 可读性更好的字符串表示形式
    """
    if n >= 1048576:
        return f"{round(n / 1048576, 2)}M"
    if n >= 1024:
        return f"{round(n / 1024, 2)}K"
    return f"{round(n, 1)}"


def is_common_text_file(url_path: str) -> bool:
    """判断是否为常见的文本文件.

    Args:
        url_path (str): 文件路径

    Returns:
        bool: 是否为常见的文本文件
    """
    return any(
        url_path.endswith(i)
        for i in [
            ".txt",
            ".yml",
            ".md",
            ".xml",
            ".html",
            ".py",
            ".h",
            ".c",
            ".pyi",
            ".js",
            ".json",
        ]
    )


def get_file_size(url: str) -> Union[int, None]:
    """获取文件大小

    Args:
        url (str): 网址

    Returns:
        Union[int, None]: 文件大小（单位：字节）
    """
    response = requests.head(url, timeout=10)
    if "Content-Length" in response.headers:
        file_size = int(response.headers["Content-Length"])
        return file_size
    return None


def download_file_chunk(url: str, start_byte: int, end_byte: int, save_dir: str) -> int:
    """下载文件的代码块

    Args:
        url (str): 文件的 URL 地址
        start_byte (int): 下载的起始字节位置
        end_byte (int): 下载的结束字节位置
        save_dir (str): 文件保存的目录

    Returns:
        int: 已下载的字节数
    """
    headers = {"Range": f"bytes={start_byte}-{end_byte}"}
    response = requests.get(url, headers=headers, stream=True, timeout=10)
    response.raise_for_status()
    with open(save_dir + ".tmp", "rb+") as dwnf:
        dwnf.seek(start_byte)
        downloaded_bytes = 0
        for chunk in response.iter_content(chunk_size=8192):
            dwnf.write(chunk)
            downloaded_bytes += len(chunk)
        return downloaded_bytes


def download_file_singlethreaded(url: str, save_dir: str) -> None:
    """下载单个文件

    Args:
        url (str): 文件的 URL 地址
        save_dir (str): 文件保存的目录
        ignore_warnings (bool, optional): 是否忽略警告
    """
    with requests.get(url, stream=True, timeout=10) as res:
        res.raise_for_status()
        filesize = get_file_size(url)
        chunk_size = 8192
        useSpeed: float = 0
        # nowsize: 当前已下载的字节数
        nowsize: int = 0
        lastime: float = 1
        with open(save_dir + ".tmp", "wb") as dwnf:
            for chk in res.iter_content(chunk_size):
                nowtime = time.time()
                if nowtime != lastime:
                    useSpeed = chunk_size / (nowtime - lastime)
                    lastime = nowtime
                nowsize += len(chk)
                dwnf.write(chk)
                if nowsize % 81920 == 0 and filesize:  # 每下载 10 个数据块更新一次进度
                    download_progress_bar(nowsize, filesize, useSpeed)
        shutil.move(save_dir + ".tmp", save_dir)


def download_unknown_file(url: str, save_dir: str) -> None:
    """下载未知文件

    Args:
        url (str): 文件的 URL 地址
        save_dir (str): 文件保存的目录
    """
    # 鉴于 Content-Length 不一定表示文件原始大小，二进制文件与文本文件需要分开下载
    # 否则显示的下载条会异常
    resp = requests.get(url, timeout=10)
    resp.raise_for_status()
    if is_common_text_file(save_dir):
        # 文本文件，体积可能不大
        with open(save_dir, "wb") as f:
            f.write(resp.content)
    download_file_singlethreaded(url, save_dir)


def test_site_latency(Da: dict) -> list:
    """测试网站延迟

    Args:
        Da (dict): 包含 URL 和镜像 URL 的字典

    Returns:
        list: 按延迟排序的 URL 和延迟时间的元组列表
    """
    tmp_speed = {}
    urls = [Da["url"]] + Da["mirror_url"]

    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(measure_latencyt, url) for url in urls]

        for future, url in zip(as_completed(futures), urls):
            try:
                latency = future.result(timeout=5)
                if latency != -1:
                    tmp_speed[url] = latency
            except Exception as e:
                Print.print_war(f"Error measuring latency for {url}: {e}")

    return sorted(tmp_speed.items(), key=lambda x: x[1], reverse=True)


def measure_latencyt(url: str) -> float:
    """测量延迟

    Args:
        url (str): 网址

    Raises:
        ValueError: 无效的网址

    Returns:
        float: 延迟时间
    """
    try:
        # 提取域名
        domain = re.search(r"(?<=http[s]://)[.\w-]*(:\d{1,8})?((?=/)|(?!/))", url)
        if isinstance(domain, type(None)):  # 如果没有匹配到域名
            raise ValueError("Invalid URL")
        st = pyspeedtest.SpeedTest(domain.group())  # 传入域名
        download_speed = st.download()
        return download_speed
    except Exception as e:
        Print.print_war(f"Error measuring latency for {url}: {e}")
    return -1.0  # 返回 -1 表示测速失败


def get_free_port(start: int = 2000, end: int = 65535) -> int:
    """获取空闲端口号

    Args:
        start (int, optional): 起始端口号.
        end (int, optional): 结束端口号.

    Raises:
        Exception: 未找到空闲端口

    Returns:
        int: 空闲端口号
    """
    for port in range(start, end):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return port
            except OSError:
                Print.print_war(f"端口 {port} 正被占用，跳过")
    raise ValueError(f"未找到空闲端口 ({start}~{end})")


def check_update() -> None:
    """检查更新"""
    try:
        latest_version: str = requests.get(
            "https://tdload.tblstudio.cn/https://api.github.com/repos/ToolDelta/ToolDelta/releases/latest",
            timeout=5,
        ).json()["tag_name"]
        current_version = ".".join(map(str, get_tool_delta_version()[:3]))

        if not latest_version.replace(".", "") <= current_version.replace(".", ""):
            # Print.print_suc(f"当前为最新版本 -> v{current_version}，无需更新")
            Print.print_load(
                f"检测到最新版本 {current_version} -> {latest_version}，请及时更新！"
            )
    except KeyError:
        Print.print_war("获取最新版本失败，请检查网络连接")
    except Exception as err:
        Print.print_war(f"无法获取最新版本: {str(err)[:30]}.., 已忽略")


def if_token() -> None:
    """检查路径下是否有 fbtoken，没有就提示输入

    Raises:
        SystemExit: 未输入 fbtoken
    """
    if not os.path.isfile("fbtoken"):
        Print.print_inf(
            "请到对应的验证服务器官网下载 FBToken，并放在本目录中，或者在下面输入 fbtoken"
        )
        fbtoken = input(Print.fmt_info("请输入 fbtoken: ", "§b 输入 "))
        if fbtoken:
            with open("fbtoken", "w", encoding="utf-8") as f:
                f.write(fbtoken)
        else:
            Print.print_err("未输入 fbtoken, 无法继续")
            raise SystemExit


def fbtokenFix():
    """修复 fbtoken 里的换行符"""
    with open("fbtoken", "r", encoding="utf-8") as f:
        token = f.read()
        if "\n" in token:
            Print.print_war("fbtoken 里有换行符，会造成 fb 登陆失败，已自动修复")
            with open("fbtoken", "w", encoding="utf-8") as f:
                f.write(token.replace("\n", ""))



"还原游戏常见字符串"

import gzip
import os
import re
import tarfile
import threading
import warnings
from glob import glob
from importlib import util
from typing import Dict

import requests
import ujson as json
import urllib3

from .color_print import Print
from .constants import TDSPECIFIC_MIRROR
from .get_tool_delta_version import get_tool_delta_version
from .sys_args import sys_args_to_dict
from .urlmethod import download_file_singlethreaded

# 关闭警告
urllib3.disable_warnings()
warnings.filterwarnings("ignore")


class GameTextsLoader:
    "还原游戏常见字符串"

    def __init__(self) -> None:
        "初始化"
        self.base_path = os.path.join(os.getcwd(), "插件数据文件", "game_texts")
        self.check_initial_run()
        if "no-download-libs" not in sys_args_to_dict():
            self.start_auto_update_thread()
            self.auto_update()
        self.game_texts_data: Dict[str, str] = self.load_data()

    @staticmethod
    def get_latest_version() -> str:
        """获取最新版本号

        Returns:
            str: 版本号
        """
        if (
            "no-download-libs" in sys_args_to_dict()
            or "no-update-check" in sys_args_to_dict()
        ):
            return ".".join(map(str, get_tool_delta_version()))
        result = re.match(
            r"(\d+\.\d+\.\d+)",
            requests.get(
                "https://tdload.tblstudio.cn/https://api.github.com/repos/ToolDelta/GameText/releases/latest",
                timeout=5,
                verify=True,
            ).json()["tag_name"],
        )
        if not isinstance(result, type(None)):
            return result.group()
        raise ValueError("无法获取最新版本号")

    def check_initial_run(self) -> None:
        "检查初始运行"
        version_file_path: str = os.path.join(self.base_path, "version")
        if not os.path.exists(version_file_path):
            latest_version: str = self.get_latest_version()
            with open(version_file_path, "w", encoding="utf-8") as f:
                f.write(latest_version)
            self.download_and_extract(latest_version)

    def start_auto_update_thread(self) -> None:
        "启用自动更新线程"
        threading.Timer(24 * 60 * 60, self.auto_update).start()

    def auto_update(self) -> None:
        "自动更新"
        version_file_path: str = os.path.join(self.base_path, "version")
        with open(version_file_path, "r", encoding="utf-8") as f:
            version: str = f.read()
        latest_version: str = self.get_latest_version()
        if version != latest_version:
            self.download_and_extract(latest_version)
            with open(version_file_path, "w", encoding="utf-8") as f:
                f.write(latest_version)

    def download_and_extract(self, version) -> None:
        "下载并解压"
        packets_url: str = (
            f"{TDSPECIFIC_MIRROR}/https://github.com/ToolDelta/"
            f"GameText/releases/download/{version}/"
            "ToolDelta_Game_Texts.tar.gz"
        )
        archive_path = os.path.join(self.base_path, "ToolDelta_Game_Texts.tar.gz")
        download_file_singlethreaded(packets_url, archive_path)
        self.extract_data_archive(archive_path)

    def load_data(self) -> Dict[str, str]:
        """加载数据

        Returns:
            Dict[str, str]: 数据
        """
        try:
            all_values: Dict[str, str] = {}
            for file_path in glob(
                pathname=os.path.join(self.base_path, "src", "**", "*.py"),
                recursive=True,
            ):
                module_name: str = os.path.basename(file_path).replace(".py", "")
                spec = util.spec_from_file_location(module_name, file_path)
                if isinstance(spec, type(None)) or isinstance(spec.loader, type(None)):
                    Print.print_war(f"Failed to load module from {file_path}")
                    continue
                module = util.module_from_spec(spec)
                spec.loader.exec_module(module)
                for var_name in dir(module):
                    if not var_name.startswith("__"):
                        var_value = getattr(module, var_name)
                        if isinstance(var_value, dict):
                            all_values.update(var_value)
            return all_values
        except Exception as err:
            print(f"Error loading game texts data: {err}")
            return {}

    def extract_data_archive(self, zip_path: str) -> bool:
        """解压数据归档

        Args:
            zip_path (str): 压缩包路径

        Returns:
            bool: 是否成功
        """
        try:
            with gzip.open(zip_path, "rb") as f_in, tarfile.open(
                fileobj=f_in,  # type: ignore
                mode="r:gz",
            ) as tar:
                tar.extractall(self.base_path)
                with open(
                    os.path.join(self.base_path, "src_tree.json"), "w", encoding="utf-8"
                ) as f:
                    json.dump(tar.getnames(), f)
                return True
        except Exception as err:
            Print.print_war(f"Error extracting data archive: {err}")
            return False


class GameTextsHandle:
    """处理游戏文本返回"""

    def __init__(self, Game_Texts: dict) -> None:
        self.Game_Texts = Game_Texts

    def Handle_Text_Class1(self, packet: dict | list) -> list:
        """处理文本返回方法 1

        Args:
            packet (dict | list): 数据包

        Returns:
            list: 处理结果的 json 格式列表
        """
        json_result: list = []  # 使用 list 保存处理后的结果
        if isinstance(packet, list):  # 如果 packet 是 list 类型
            for item in packet:  # 遍历 packet 中的每个 item
                # 从 self.Game_Texts 中获取原始消息文本
                if not isinstance(
                    (
                        original_message := self.Game_Texts.get(
                            item["Message"].replace("%", "")
                        )
                    ),
                    type(None),
                ):
                    # 检查原始消息中是否包含格式化参数%
                    if not len(re.findall(r"%[a-zA-Z]", original_message)) >= 1:
                        if original_message:  # 如果原始消息不为空
                            original_message = re.sub(
                                r'\$[^"\'\]/\]\)）}\s]{0,3}', "", original_message
                            )  # 删除$后的内容
                            param_list = list(item["Parameters"])  # 获取参数列表
                            for n, _ in enumerate(param_list, start=1):
                                original_message = original_message.replace(
                                    f"{n}", f"{{{n - 1}}}"
                                )
                            # 检查参数中是否包含%
                            if (
                                len(
                                    [
                                        str(param)
                                        for param in param_list
                                        if "%" in str(param)
                                    ]
                                )
                                >= 1
                            ):
                                # 提取参数中的%后的字符并替换为实际的值
                                filtered_param_list = [
                                    re.sub(r"%", "", param)
                                    for param in param_list
                                    if "%" in param
                                ]
                                for filtered_param in filtered_param_list:
                                    for i in range(len(param_list)):
                                        if filtered_param in param_list[i]:
                                            param_list[i] = param_list[i].replace(
                                                f"%{filtered_param}",
                                                self.Game_Texts.get(
                                                    filtered_param, "???"
                                                ),
                                            )
                            # 格式化消息文本
                            filled_message = original_message.format(*param_list)
                    else:
                        if original_message:  # 如果原始消息不为空
                            original_message = re.sub(
                                r'\$[^"\'\]/\]\)）}\s]{0,3}', "", original_message
                            )  # 删除$后的内容
                            param_list = list(item["Parameters"])  # 获取参数列表
                            re.findall(r"%[a-zA-Z]", original_message)  # 查找格式化参数
                            formatted_string = original_message
                            for i, arg in enumerate(param_list, start=0):
                                formatted_string = re.sub(
                                    r"%[a-zA-Z]", str(arg), formatted_string, count=1
                                )
                            if (
                                len(
                                    [
                                        str(param)
                                        for param in param_list
                                        if "%" in str(param)
                                    ]
                                )
                                >= 1
                            ):
                                # 提取参数中的%后的字符并替换为实际的值
                                filtered_param_list = [
                                    re.sub(r"%", "", param)
                                    for param in param_list
                                    if "%" in param
                                ]
                                for filtered_param in filtered_param_list:
                                    for i in range(len(param_list)):
                                        if filtered_param in param_list[i]:
                                            param_list[i] = param_list[i].replace(
                                                f"%{filtered_param}",
                                                self.Game_Texts.get(filtered_param),
                                            )
                            # 格式化消息文本
                            filled_message = formatted_string.format(*param_list)
                else:
                    filled_message = item["Message"]
                # 将填充后的消息文本转换为 json 格式，并添加到结果列表中
                json_output = json.dumps(filled_message, indent=2, ensure_ascii=False)
                json_result.append(json_output)
        elif isinstance(packet, dict):  # 如果 packet 不是 list 类型
            # 从 self.Game_Texts 中获取原始消息文本
            if not isinstance(
                (
                    original_message := self.Game_Texts.get(
                        packet["Message"].replace("%", "")
                    )
                ),
                type(None),
            ):
                if not len(re.findall(r"%[a-zA-Z]", original_message)) >= 1:
                    if original_message:  # 如果原始消息不为空
                        original_message = re.sub(
                            r'\$[^"\'\]/\]\)）}\s]{0,3}', "", original_message
                        )  # 删除$后的内容
                        param_list = list(packet["Parameters"])  # 获取参数列表
                        for n, _ in enumerate(param_list, start=1):
                            original_message = original_message.replace(
                                f"%{n}", "{" + str(n - 1) + "}"
                            )
                        if (
                            len(
                                [
                                    str(param)
                                    for param in param_list
                                    if "%" in str(param)
                                ]
                            )
                            >= 1
                        ):
                            # 提取参数中的%后的字符并替换为实际的值
                            filtered_param_list = [
                                re.sub(r"%", "", param)
                                for param in param_list
                                if "%" in param
                            ]
                            for filtered_param in filtered_param_list:
                                for i in range(len(param_list)):
                                    if filtered_param in param_list[i]:
                                        param_list[i] = param_list[i].replace(
                                            f"%{filtered_param}",
                                            self.Game_Texts.get(filtered_param),
                                        )
                        # 格式化消息文本
                        filled_message = original_message.format(*param_list)
                else:
                    if original_message := self.Game_Texts.get(
                        packet["Message"].replace("%", "")
                    ):  # 如果原始消息不为空
                        param_list = list(packet["Parameters"])  # 获取参数列表
                        original_message = re.sub(
                            r'\$[^"\'\]/\]\)）}\s]{0,3}', "", original_message
                        )  # 删除$后的内容
                        formatted_string = original_message
                        for i, arg in enumerate(param_list, start=1):
                            formatted_string = re.sub(
                                r"%[a-zA-Z]", str(arg), formatted_string, count=1
                            )
                        if (
                            len(
                                [
                                    str(param)
                                    for param in param_list
                                    if "%" in str(param)
                                ]
                            )
                            >= 1
                        ):
                            # 提取参数中的%后的字符并替换为实际的值
                            filtered_param_list = [
                                re.sub(r"%", "", param)
                                for param in param_list
                                if "%" in param
                            ]
                            for filtered_param in filtered_param_list:
                                for i in range(len(param_list)):
                                    if filtered_param in param_list[i]:
                                        param_list[i] = param_list[i].replace(
                                            f"%{filtered_param}",
                                            self.Game_Texts.get(filtered_param),
                                        )

                        filled_message = formatted_string
            else:
                filled_message = packet["Message"]
            # 将填充后的消息文本转换为 json 格式，并添加到结果列表中
            json_output = json.dumps(filled_message, indent=2, ensure_ascii=False)
            json_result.append(json_output)
        return json_result  # 返回处理结果的 json 格式列表



"""提供了一些实用方法的类"""

import copy
import ctypes
import json as rjson
import os
import threading
import time
import traceback
from io import TextIOWrapper
from typing import Any, Callable, Iterable, Optional, TypeVar

import ujson as json

from .color_print import Print
from .constants import TOOLDELTA_PLUGIN_DATA_DIR

event_pool = {"timer_events": threading.Event()}
threads_list: list["Utils.createThread"] = []
timer_events_table: dict[int, tuple[str, Callable, tuple, dict]] = {}

VT = TypeVar("VT")


class Utils:
    """提供了一些实用方法的类"""

    class ThreadExit(SystemExit):
        """线程退出."""

    class ToolDeltaThread(threading.Thread):
        """简化 ToolDelta 子线程创建的 threading.Thread 的子类."""

        def __init__(
            self, func: Callable, args: Iterable[Any] = (), usage="", **kwargs
        ):
            """新建一个 ToolDelta 子线程

            Args:
                func (Callable): 线程方法
                args (tuple, optional): 方法的参数项
                usage (str, optional): 线程的用途说明
                kwargs (dict, optional): 方法的关键词参数项
            """
            super().__init__(target=func)
            self.func = func
            self.daemon = True
            self.all_args = [args, kwargs]
            self.usage = usage
            self.start()
            self.stopping = False
            self._thread_id = None

        def run(self) -> None:
            """线程运行方法"""
            threads_list.append(self)
            try:
                self.func(*self.all_args[0], **self.all_args[1])
            except SystemExit:
                pass
            except ValueError as e:
                if str(e) != "未连接到游戏":
                    raise
                Print.print_war(f"线程 {self.usage} 因游戏断开连接被迫中断")
            except Exception:
                Print.print_err(
                    f"线程 {self.usage or self.func.__name__} 出错:\n"
                    + traceback.format_exc()
                )
            finally:
                threads_list.remove(self)

        def get_id(self) -> int:
            """获取线程的 ID

            Raises:
                RuntimeError: 线程 ID 未知

            Returns:
                int: 线程 ID
            """
            if self._thread_id is None:
                for thread in threading.enumerate():
                    if thread is self:
                        self._thread_id = thread.ident
                        break
            if self._thread_id is None:
                raise RuntimeError("Could not determine the thread's ID")
            return self._thread_id

        def stop(self) -> None:
            """终止线程"""
            self.stopping = True
            self._thread_id = self.ident
            thread_id = self.get_id()
            res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
                thread_id, ctypes.py_object(SystemExit)
            )
            if res > 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
                Print.print_err(f"§c终止线程 {self.name} 失败")

    createThread = ClassicThread = ToolDeltaThread

    class TMPJson:
        """提供了加载、卸载、读取和写入 JSON 文件到缓存区的方法的类."""

        @staticmethod
        def loadPathJson(path: str, needFileExists: bool = True) -> None:
            """
            将 json 文件从磁盘加载到缓存区，以便快速读写.
            在缓存文件已加载的情况下，再使用一次该方法不会有任何作用.

            Args:
                path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
                needFileExists (bool, optional): 默认为 True, 为 False 时，若文件路径不存在，就会自动创建一个文件，且默认值为 null

            Raises:
                err: 文件不存在时
            """
            if path in jsonPathTmp:
                return
            try:
                with open(path, "r", encoding="utf-8") as file:
                    js = Utils.SimpleJsonDataReader.SafeJsonLoad(file)
            except FileNotFoundError as err:
                if not needFileExists:
                    js = None
                else:
                    raise err from None
            jsonPathTmp[path] = [False, js]

        @staticmethod
        def unloadPathJson(path: str) -> bool:
            """
            将 json 文件从缓存区卸载 (保存内容到磁盘), 之后不能再在缓存区对这个文件进行读写.
            在缓存文件已卸载的情况下，再使用一次该方法不会有任何作用，但是可以通过其返回的值来知道存盘有没有成功.

            Args:
                path (str): 文件的虚拟路径

            Returns:
                bool: 存盘是否成功
            """
            if jsonPathTmp.get(path) is not None:
                isChanged, dat = jsonPathTmp[path]
                if isChanged:
                    with open(path, "w", encoding="utf-8") as file:
                        Utils.SimpleJsonDataReader.SafeJsonDump(dat, file)
                del jsonPathTmp[path]
                return True
            return False

        @staticmethod
        def read(path: str) -> Any:
            """对缓存区的该虚拟路径的文件进行读操作，返回一个深拷贝的 JSON 对象

            Args:
                path (str): 文件的虚拟路径

            Raises:
                Exception: json 路径未初始化，不能进行读取和写入操作

            Returns:
                list[Any] | dict[Any, Any] | Any: 该虚拟路径的 JSON
            """
            if path in jsonPathTmp:
                val = jsonPathTmp.get(path)
                if val is not None:
                    val = val[1]
                    if isinstance(val, (list, dict)):
                        val = copy.deepcopy(val)
                    return val
            raise ValueError("json 路径未初始化，不能进行读取和写入操作：" + path)

        @staticmethod
        def get(path: str) -> None:
            """
            直接获取缓存区的该虚拟路径的 JSON, 不使用 copy
            WARNING: 如果你不知道有什么后果，请老老实实使用`read(...)`而不是`get(...)`!

            Args:
                path (str): 文件的虚拟路径
            """
            if path in jsonPathTmp:
                val = jsonPathTmp.get(path)
                if val is not None:
                    val = val[1]
                    return val
            raise ValueError("json 路径未初始化，不能进行读取和写入操作：" + path)

        @staticmethod
        def write(path: str, obj: Any) -> None:
            """
            对缓存区的该虚拟路径的文件进行写操作，这将会覆盖之前的内容

            Args:
                path (str): 文件的虚拟路径
                obj (Any): 任何合法的 JSON 类型 例如 dict/list/str/bool/int/float
            """
            if path in jsonPathTmp:
                jsonPathTmp[path] = [True, obj]
            else:
                raise ValueError("json 路径未初始化，不能进行读取和写入操作：" + path)

        @staticmethod
        def read_as_tmp(
            path: str, needFileExists: bool = True, timeout: int = 60
        ) -> Any:
            """读取 json 文件并将其从磁盘加载到缓存区，以便一段时间内能快速读写.

            Args:
                path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
                needFileExists (bool, optional): 默认为 True, 为 False 时，若文件路径不存在，就会自动创建一个文件，且写入默认值 null
                timeout (int, optional): 多久没有再进行读取操作时卸载缓存

            Returns:
                Any: 该虚拟路径的 JSON
            """
            if path not in jsonUnloadPathTmp and path not in jsonPathTmp:
                jsonUnloadPathTmp[path] = timeout + int(time.time())
                Utils.TMPJson.loadPathJson(path, needFileExists)
            return Utils.TMPJson.read(path)

        @staticmethod
        def write_as_tmp(
            path: str, obj: Any, needFileExists: bool = True, timeout: int = 60
        ) -> None:
            """写入 json 文件并将其从磁盘加载到缓存区，以便一段时间内能快速读写.

            Args:
                path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
                obj (Any): 任何合法的 JSON 类型 例如 dict/list/str/bool/int/float
                needFileExists (bool, optional): 默认为 True, 为 False 时，若文件路径不存在，就会自动创建一个文件，且写入默认值 null
                timeout (int, optional): 多久没有再进行读取操作时卸载缓存
            """
            if path not in jsonUnloadPathTmp and path not in jsonPathTmp:
                jsonUnloadPathTmp[path] = timeout + int(time.time())
                Utils.TMPJson.loadPathJson(path, needFileExists)
            Utils.TMPJson.write(path, obj)

        @staticmethod
        def cancel_change(path: str) -> None:
            """取消缓存 json 所做的更改，非必要情况请勿调用，你不知道什么时候会自动保存所做更改"""
            jsonPathTmp[path][0] = False

        @staticmethod
        def get_tmps() -> dict:
            """不要调用!"""
            return jsonPathTmp.copy()

    class JsonIO:
        """提供了安全的 JSON 文件操作方法的类."""

        @staticmethod
        def SafeJsonDump(obj: Any, fp: TextIOWrapper | str, indent=2) -> None:
            """将一个 json 对象写入一个文件，会自动关闭文件读写接口.

            Args:
                obj (str | dict | list): JSON 对象
                fp (Any): open(...) 打开的文件读写口 或 文件路径
            """
            if isinstance(fp, str):
                with open(fp, "w") as file:
                    file.write(json.dumps(obj, indent=indent, ensure_ascii=False))
            else:
                with fp:
                    fp.write(json.dumps(obj, indent=indent, ensure_ascii=False))

        @staticmethod
        def SafeJsonLoad(fp: TextIOWrapper | str) -> Any:
            """从一个文件读取 json 对象，会自动关闭文件读写接口.

            Args:
                fp (TextIOWrapper | str): open(...) 打开的文件读写口 或文件路径

            Returns:
                dict | list: JSON 对象
            """
            if isinstance(fp, str):
                with open(fp, "r", encoding="utf-8") as file:
                    return json.load(file)
            with fp as file:
                return json.load(file)

        class DataReadError(json.JSONDecodeError):
            """读取数据时发生错误"""

        @staticmethod
        def readFileFrom(
            plugin_name: str, file: str, default: dict | None = None
        ) -> Any:
            """从插件数据文件夹读取一个 json 文件，会自动创建文件夹和文件.

            Args:
                plugin_name (str): 插件名
                file (str): 文件名
                default (dict, optional): 默认值，若文件不存在则会写入这个默认值

            Raises:
                Utils.JsonIO.DataReadError: 读取数据时发生错误
                err: 读取文件路径时发生错误

            Returns:
                dict | list: JSON 对象
            """
            if file.endswith(".json"):
                file = file[:-5]
            filepath = os.path.join(
                TOOLDELTA_PLUGIN_DATA_DIR, plugin_name, f"{file}.json"
            )
            os.makedirs(
                os.path.join(TOOLDELTA_PLUGIN_DATA_DIR, plugin_name), exist_ok=True
            )
            try:
                if default is not None and not os.path.isfile(filepath):
                    with open(filepath, "w") as f:
                        Utils.JsonIO.SafeJsonDump(default, f)
                    return default
                with open(filepath, "r", encoding="utf-8") as f:
                    res = Utils.JsonIO.SafeJsonLoad(f)
                return res
            except rjson.JSONDecodeError as err:
                # 判断是否有 msg.doc.pos 属性
                raise Utils.JsonIO.DataReadError(err.msg, err.doc, err.pos)
            except Exception as err:
                Print.print_err(f"读文件路径 {filepath} 发生错误")
                raise err

        @staticmethod
        def writeFileTo(plugin_name: str, file: str, obj: Any, indent=4) -> None:
            """将一个 json 对象写入插件数据文件夹，会自动创建文件夹和文件.

            Args:
                plugin_name (str): 插件名
                file (str): 文件名
                obj (str | dict[Any, Any] | list[Any]): JSON 对象
            """
            os.makedirs(f"{TOOLDELTA_PLUGIN_DATA_DIR}/{plugin_name}", exist_ok=True)
            with open(
                f"{TOOLDELTA_PLUGIN_DATA_DIR}/{plugin_name}/{file}.json", "w"
            ) as f:
                Utils.JsonIO.SafeJsonDump(obj, f, indent=indent)

    SimpleJsonDataReader = JsonIO

    class ChatbarLock:
        """
        聊天栏锁, 用于防止玩家同时开启多个聊天栏对话

        调用了该锁的所有代码, 在另一个进程使用该锁的时候, 尝试调用其他锁会导致进程直接退出, 直到此锁退出为止

        示例(以类式插件为例):
        ```python
        class MyPlugin(Plugin):
            ...
            def on_player_message(self, player: str, msg: str):
                with ChatbarLock(player):
                    # 如果玩家处在另一个on_player_message进程 (锁环境) 中
                    # 则在上面就会直接引发 SystemExit
                    ...
        ```
        示例(以注入式插件为例):
        ```
        @player_message()
        async def onPlayerChat(info: player_message_info):
            with ChatbarLock(info.playername):
                ...
        ```
        """

        def __init__(self, player: str, oth_cb: Callable[[str], None] = lambda _: None):
            self.player = player
            self.oth_cb = oth_cb

        def __enter__(self):
            if self.player in chatbar_lock_list:
                self.oth_cb(self.player)
                Print.print_war(f"玩家 {self.player} 的线程锁正在锁定状态")
                raise SystemExit
            chatbar_lock_list.append(self.player)

        def __exit__(self, e, e2, e3):
            chatbar_lock_list.remove(self.player)

    @staticmethod
    def get_threads_list() -> list["Utils.createThread"]:
        """返回使用 createThread 创建的全线程列表。"""
        return threads_list

    @staticmethod
    def simple_fmt(kw: dict[str, Any], sub: str) -> str:
        """
        快速将字符串内按照给出的 dict 的键值对替换掉内容.

        参数:
            kw: Dict[str, Any], 键值对应替换的内容
            *args: str, 需要被替换的字符串

        示例:
            >>> my_color = "red"; my_item = "apple"
            >>> kw = {"[颜色]": my_color, "[物品]": my_item}
            >>> Utils.SimpleFmt(kw, "I like [颜色] [物品].")
            I like red apple.
        """
        for k, v in kw.items():
            if k in sub:
                sub = sub.replace(k, str(v))
        return sub

    SimpleFmt = simple_fmt

    @staticmethod
    def simple_assert(cond: Any, exc: Any) -> None:
        """相当于 assert cond, 但是可以自定义引发的异常的类型"""
        if not cond:
            raise exc

    simpleAssert = simple_assert

    @staticmethod
    def thread_func(func_or_name: Callable | str) -> Any:
        """
        在事件方法可能执行较久会造成堵塞时使用，方便快捷地创建一个新线程，例如:

        ```python
        @Utils.thread_func
        def on_inject(self):
            ...
        ```
        或者:
        ```python
        @Utils.thread_func("一个会卡一分钟的线程")
        def on_inject(self):
            ...
        ```
        """
        if isinstance(func_or_name, str):

            def _recv_func(func: Callable):
                def thread_fun(*args: tuple, **kwargs: Any) -> None:
                    Utils.createThread(func, usage=func_or_name, args=args, **kwargs)

                return thread_fun

            return _recv_func

        def thread_fun(*args: tuple, **kwargs: Any) -> None:
            Utils.createThread(
                func_or_name,
                usage="简易线程方法：" + func_or_name.__name__,
                args=args,
                **kwargs,
            )

        return thread_fun

    @staticmethod
    def timer_event(t: int, name: Optional[str] = None):
        """
        将修饰器下的方法作为一个定时任务, 每隔一段时间被执行一次。
        注意: 请不要在函数内放可能造成堵塞的内容
        注意: 当此函数被修饰后, 需要调用一次才能开始定时任务线程!

        Args:
            seconds (int): 周期秒数
            name (Optional[str], optional): 名字, 默认为自动生成的
        """

        def receiver(func: Callable[[], None] | Callable[[Any], None]):
            def caller(*args, **kwargs):
                func_name = name or f"简易方法:{func.__name__}"
                timer_events_table[t] = (func_name, func, args, kwargs)

            return caller

        return receiver

    @staticmethod
    def try_int(arg: Any) -> Optional[int]:
        """尝试将提供的参数化为 int 类型并返回，否则返回 None"""
        try:
            return int(arg)
        except Exception:
            return None

    @staticmethod
    def fuzzy_match(lst: list[str], sub: str) -> list[str]:
        """
        模糊匹配列表内的字符串，可以用在诸如模糊匹配玩家名的用途

        参数:
            lst: list, 字符串列表
            sub: str, 需要匹配的字符串
        返回:
            list, 匹配结果
        """
        res = []
        for i in lst:
            if sub in i:
                res.append(i)
        return res

    @staticmethod
    def split_list(lst: list[VT], length: int) -> list[list[VT]]:
        """
        将列表进行块分割

        Args:
            lst (list[VT]): 传入列表
            length (int): 分割的单个列表的长度

        Returns:
            list[list[VT]]: 传出的被分割的列表
        """
        return [lst[i : i + length] for i in range(0, len(lst), length)]


def safe_close() -> None:
    """安全关闭"""
    event_pool["timer_events"].set()
    print(jsonPathTmp)
    _tmpjson_save()


def _tmpjson_save():
    "请不要在系统调用以外调用"
    for k, (isChanged, dat) in jsonPathTmp.copy().items():
        if isChanged:
            Utils.SimpleJsonDataReader.SafeJsonDump(dat, k)
            jsonPathTmp[k][0] = False
    for k, v in jsonUnloadPathTmp.copy().items():
        if time.time() - v > 0:
            Utils.TMPJson.unloadPathJson(k)
            del jsonUnloadPathTmp[k]


@Utils.timer_event(1, "缓存JSON数据定时保存")
@Utils.thread_func("JSON 缓存文件定时保存")
def tmpjson_save():
    "请不要在系统调用以外调用"
    _tmpjson_save()


@Utils.thread_func("ToolDelta 定时任务")
def timer_event_boostrap():
    "请不要在系统调用以外调用"
    timer = 0
    evt = event_pool["timer_events"]
    while not evt.is_set():
        for k, (_, v, a, kwa) in timer_events_table.items():
            if timer % k == 0:
                v(*a, **kwa)
        evt.wait(1)
        timer += 1


jsonPathTmp = {}
chatbar_lock_list = []
jsonUnloadPathTmp = {}



"""ToolDelta 启动器"""

import os
import time
import traceback

from .color_print import Print
from .frame import GameCtrl, ToolDelta
from .plugin_load.PluginGroup import plugin_group
from .sys_args import sys_args_to_dict
from .urlmethod import check_update
from .utils import timer_event_boostrap, tmpjson_save

tooldelta = ToolDelta()


def start_tool_delta() -> None:
    """启动 ToolDelta"""
    try:
        tooldelta.welcome()
        if "no-update-check" not in sys_args_to_dict():
            check_update()
        else:
            Print.print_war("将不会进行自动更新。")
        tooldelta.basic_operation()
        tooldelta.loadConfiguration()
        tooldelta.launcher.init()
        game_control = GameCtrl(tooldelta)
        tooldelta.set_game_control(game_control)
        tooldelta.set_plugin_group(plugin_group)
        plugin_group.set_frame(tooldelta)
        plugin_group.read_all_plugins()
        timer_event_boostrap()
        tmpjson_save()
        tooldelta.launcher.listen_launched(game_control.Inject)
        game_control.set_listen_packets()
        raise tooldelta.launcher.launch()
    except (KeyboardInterrupt, SystemExit, EOFError):
        Print.print_inf("ToolDelta 已关闭")
        pass
    except Exception:
        Print.print_err(f"ToolDelta 运行过程中出现问题：{traceback.format_exc()}")
        input(Print.clean_fmt("§c按回车键退出..."))


def safe_jump(out_task: bool = True, exit_directly: bool = True) -> None:
    """安全退出

    Args:
        out_task (bool, optional): frame 框架系统是否退出
        exit_directly (bool, optional): 是否三秒强制直接退出
    """
    if out_task:
        tooldelta.system_exit()
    tooldelta.safelyExit()
    if exit_directly:
        for _ in range(3, -1, -1):
            Print.print_war(f"{_}秒后强制退出...", end="\r")
            time.sleep(1)
        Print.print_suc("ToolDelta 已退出。")
        os._exit(0)
    Print.print_suc("ToolDelta 已退出。")



"""
ToolDelta 基本框架

整个系统由三个部分组成
    Frame: 负责整个 ToolDelta 的基本框架运行
    GameCtrl: 负责对接游戏
        - Launchers: 负责将不同启动器的游戏接口统一成固定的接口，供插件在多平台游戏接口运行 (FastBuilder External, NeOmega, (BEWS, etc.))
    PluginGroup: 负责对插件进行统一管理
"""

import asyncio
import contextlib
import getpass
import os
import signal
import sys
import time
import traceback
from typing import TYPE_CHECKING, Callable

import requests
import ujson as json

from tooldelta import (
    auths,
    constants,
    plugin_market,
)

from .cfg import Config
from .color_print import Print
from .game_texts import GameTextsHandle, GameTextsLoader
from .game_utils import getPosXYZ
from .get_tool_delta_version import get_tool_delta_version
from .launch_cli import (
    FrameBEConnect,
    FrameNeOmg,
    FrameNeOmgRemote,
    SysStatus,
)
from .logger import publicLogger
from .packets import Packet_CommandOutput, PacketIDS
from .plugin_load.injected_plugin import safe_jump
from .sys_args import sys_args_to_dict
from .urlmethod import fbtokenFix, if_token
from .utils import Utils, safe_close

sys_args_dict = sys_args_to_dict(sys.argv)
VERSION = get_tool_delta_version()

if TYPE_CHECKING:
    from .plugin_load.PluginGroup import PluginGroup

LAUNCHERS: list[tuple[str, type[FrameNeOmg | FrameNeOmgRemote]]] = [
    ("NeOmega 框架 (NeOmega 模式，租赁服适应性强，推荐)", FrameNeOmg),
    (
        "NeOmega 框架 (NeOmega 连接模式，需要先启动对应的 neOmega 接入点)",
        FrameNeOmgRemote,
    ),
]


class ToolDelta:
    """ToolDelta 主框架"""

    class FrameBasic:
        """系统基本信息"""

        system_version = VERSION
        data_path = "插件数据文件/"

    class SystemVersionException(SystemError):
        """系统版本异常"""

        msg: str

    def __init__(self) -> None:
        """初始化"""
        self.createThread = Utils.createThread
        self.sys_data = self.FrameBasic()
        self.launchMode: int = 0
        self.consoleMenu = []
        self.is_docker: bool = os.path.exists("/.dockerenv")
        self.on_plugin_err = staticmethod(
            lambda name, _, err: Print.print_err(f"插件 <{name}> 出现问题：\n{err}")
        )
        self.launcher: FrameNeOmg | FrameNeOmgRemote
        self.is_mir: bool
        self.plugin_market_url: str
        self.link_game_ctrl: "GameCtrl"
        self.link_plugin_group: "PluginGroup"

    def loadConfiguration(self) -> None:
        """加载配置文件"""
        Config.default_cfg("ToolDelta基本配置.json", constants.LAUNCH_CFG)
        try:
            # 读取配置文件
            cfgs = Config.get_cfg("ToolDelta基本配置.json", constants.LAUNCH_CFG_STD)
            self.launchMode = cfgs["启动器启动模式(请不要手动更改此项, 改为0可重置)"]
            self.is_mir = cfgs["是否使用github镜像"]
            self.plugin_market_url = cfgs["插件市场源"]
            publicLogger.switch_logger(cfgs["是否记录日志"])
            if self.launchMode != 0 and self.launchMode not in range(
                1, len(LAUNCHERS) + 1
            ):
                raise Config.ConfigError(
                    "你不该随意修改启动器模式，现在赶紧把它改回 0 吧"
                )
        except Config.ConfigError as err:
            # 配置文件有误
            r = self.upgrade_cfg()
            if r:
                Print.print_war("配置文件未升级，已自动升级，请重启 ToolDelta")
            else:
                Print.print_err(f"ToolDelta 基本配置有误，需要更正：{err}")
            raise SystemExit from err
        # 每个启动器框架的单独启动配置之前
        if self.launchMode == 0:
            Print.print_inf("请选择启动器启动模式 (之后可在 ToolDelta 启动配置更改):")
            for i, (launcher_name, _) in enumerate(LAUNCHERS):
                Print.print_inf(f" {i + 1} - {launcher_name}")
            while 1:
                try:
                    ch = int(input(Print.fmt_info("请选择：", "§f 输入 ")))
                    if ch not in range(1, len(LAUNCHERS) + 1):
                        raise ValueError
                    cfgs["启动器启动模式(请不要手动更改此项, 改为0可重置)"] = ch
                    break
                except ValueError:
                    Print.print_err("输入不合法，或者是不在范围内，请重新输入")
            Config.default_cfg("ToolDelta基本配置.json", cfgs, True)
        self.launcher = LAUNCHERS[
            cfgs["启动器启动模式(请不要手动更改此项, 改为0可重置)"] - 1
        ][1]()
        # 每个启动器框架的单独启动配置
        if type(self.launcher) is FrameNeOmg:
            launch_data = cfgs.get(
                "NeOmega启动模式", constants.LAUNCHER_NEOMEGA_DEFAULT
            )
            try:
                Config.check_auto(constants.LAUNCHER_NEOMEGA_STD, launch_data)
            except Config.ConfigError as err:
                Print.print_err(
                    f"ToolDelta 基本配置-NeOmega 启动配置有误，需要更正：{err}"
                )
                raise SystemExit from err
            serverNumber = launch_data["服务器号"]
            serverPasswd = launch_data["密码"]
            auth_server = launch_data.get("验证服务器地址(更换时记得更改fbtoken)", "")
            if serverNumber == 0:
                while 1:
                    try:
                        serverNumber = int(
                            input(Print.fmt_info("请输入租赁服号：", "§b 输入 "))
                        )
                        serverPasswd = (
                            getpass.getpass(
                                Print.fmt_info(
                                    "请输入租赁服密码 (不会回显，没有请直接回车): ",
                                    "§b 输入 ",
                                )
                            )
                            or "0"
                        )
                        launch_data["服务器号"] = int(serverNumber)
                        launch_data["密码"] = int(serverPasswd)
                        cfgs["NeOmega启动模式"] = launch_data
                        Config.default_cfg("ToolDelta基本配置.json", cfgs, True)
                        Print.print_suc("登录配置设置成功")
                        break
                    except Exception:
                        Print.print_err("输入有误，租赁服号和密码应当是纯数字")
            auth_servers = constants.AUTH_SERVERS
            if auth_server == "":
                Print.print_inf("选择 ToolDelta 机器人账号 使用的验证服务器：")
                for i, (auth_server_name, _) in enumerate(auth_servers):
                    Print.print_inf(f" {i + 1} - {auth_server_name}")
                Print.print_inf(
                    "§cNOTE: 使用的机器人账号是在哪里获取的就选择哪一个验证服务器，不能混用"
                )
                while 1:
                    try:
                        ch = int(input(Print.fmt_info("请选择：", "§f 输入 ")))
                        if ch not in range(1, len(auth_servers) + 1):
                            raise ValueError
                        auth_server = auth_servers[ch - 1][1]
                        cfgs["NeOmega启动模式"][
                            "验证服务器地址(更换时记得更改fbtoken)"
                        ] = auth_server
                        break
                    except ValueError:
                        Print.print_err("输入不合法，或者是不在范围内，请重新输入")
                Config.default_cfg("ToolDelta基本配置.json", cfgs, True)
                # 读取启动配置等
            if not os.path.isfile("fbtoken"):
                Print.print_inf(
                    "请选择登录方法:\n 1 - 使用账号密码 (登录成功后将自动获取 Token 到工作目录)\n 2 - 使用 Token(如果 Token 文件不存在则需要输入或将文件放入工作目录)\r"
                )
                Login_method: str = input(
                    Print.fmt_info("请输入你的选择：", "§6 输入 ")
                )
                while True:
                    if Login_method.isdigit() is False:
                        Login_method = input(
                            Print.fmt_info("输入有误，请输入正确的序号：", "§6 警告 ")
                        )
                    elif int(Login_method) > 2 or int(Login_method) < 1:
                        Login_method = input(
                            Print.fmt_info("输入有误，请输入正确的序号：", "§6 警告 ")
                        )
                    else:
                        break
                if Login_method == "1":
                    try:
                        match cfgs["NeOmega启动模式"][
                            "验证服务器地址(更换时记得更改fbtoken)"
                        ]:
                            case "https://liliya233.uk":
                                token = auths.liliya_login()
                            case "https://api.fastbuilder.pro":
                                token = auths.fbuc_login()
                            case _:
                                Print.print_err("暂无法登录该验证服务器")
                                raise SystemExit
                        with open("fbtoken", "w", encoding="utf-8") as f:
                            f.write(token)
                    except requests.exceptions.RequestException as e:
                        Print.print_err(f"登录失败，原因：{e}\n正在切换至 Token 登录")
            if_token()
            fbtokenFix()
            with open("fbtoken", "r", encoding="utf-8") as f:
                fbtoken = f.read()
            self.launcher.set_launch_data(
                serverNumber, serverPasswd, fbtoken, auth_server
            )
        elif type(self.launcher) is FrameNeOmgRemote:
            ...
        elif type(self.launcher) is FrameBEConnect:
            launch_data = cfgs.get(
                "基岩版WS服务器启动模式", constants.LAUNCHER_BEWS_DEFAULT
            )
            try:
                Config.check_auto(constants.LAUNCHER_NEOMEGA_STD, launch_data)
            except Config.ConfigError as err:
                Print.print_err(
                    f"ToolDelta 基本配置-BEWS 启动配置有误，需要更正：{err}"
                )
                raise SystemExit from err
            if launch_data["服务端开放地址"] == "":
                Print.print_inf("请输入 WS 服务器开放的地址：")
                addr = input(
                    Print.fmt_info("请输入 (回车默认 localhost:12003): ", "§6 输入 ")
                )
                if not addr.startswith("ws://"):
                    addr = "ws://" + addr
                launch_data["服务端开放地址"] = addr
                Config.default_cfg("ToolDelta基本配置.json", cfgs, True)
        else:
            raise ValueError("LAUNCHER Error")
        Print.print_suc("配置文件读取完成")

    @staticmethod
    def change_config():
        """修改配置文件"""
        try:
            old_cfg = Config.get_cfg("ToolDelta基本配置.json", constants.LAUNCH_CFG_STD)
        except FileNotFoundError:
            Print.clean_print("§c未初始化配置文件, 无法进行修改")
            return
        except Config.ConfigError as err:
            Print.print_err(f"配置文件损坏：{err}")
            return
        if (
            old_cfg["启动器启动模式(请不要手动更改此项, 改为0可重置)"] - 1
        ) not in range(0, 2):
            Print.print_err(
                f"配置文件损坏：启动模式错误：{old_cfg['启动器启动模式(请不要手动更改此项, 改为0可重置)'] - 1}"
            )
            return
        while 1:
            md = (
                "NeOmega 框架 (NeOmega 模式，租赁服适应性强，推荐)",
                "NeOmega 框架 (NeOmega 连接模式，需要先启动对应的 neOmega 接入点)",
            )
            Print.clean_print("§b现有配置项如下:")
            Print.clean_print(
                f" 1. 启动器启动模式：{md[old_cfg['启动器启动模式(请不要手动更改此项, 改为0可重置)'] - 1]}"
            )
            Print.clean_print(f" 2. 是否记录日志：{old_cfg['是否记录日志']}")
            Print.clean_print("    §a直接回车: 保存并退出")
            resp = input(Print.clean_fmt("§6输入序号可修改配置项(0~4): ")).strip()
            if resp == "":
                Config.default_cfg("ToolDelta基本配置.json", old_cfg, True)
                Print.clean_print("§a配置已保存!")
                return
            match resp:
                case "1":
                    Print.print_inf(
                        "选择启动器启动模式 (之后可在 ToolDelta 启动配置更改):"
                    )
                    for i, (launcher_name, _) in enumerate(LAUNCHERS):
                        Print.print_inf(f" {i + 1} - {launcher_name}")
                    while 1:
                        try:
                            ch = int(input(Print.clean_fmt("请选择：")))
                            if ch not in range(1, len(LAUNCHERS) + 1):
                                raise ValueError
                            old_cfg[
                                "启动器启动模式(请不要手动更改此项, 改为0可重置)"
                            ] = ch
                            break
                        except ValueError:
                            Print.print_err("输入不合法，或者是不在范围内，请重新输入")
                            continue
                    input(
                        Print.clean_fmt(
                            f"§a已选择启动器启动模式：§f{md[old_cfg['启动器启动模式(请不要手动更改此项, 改为0可重置)'] - 1]}, 回车键继续"
                        )
                    )
                case "2":
                    old_cfg["是否记录日志"] = [True, False][old_cfg["是否记录日志"]]
                    input(
                        Print.clean_fmt(
                            f"日志记录模式已改为：{['§c关闭', '§a开启'][old_cfg['是否记录日志']]}, 回车键继续"
                        )
                    )

    @staticmethod
    def upgrade_cfg() -> bool:
        """升级配置文件

        Returns:
            bool: 是否升级了配置文件
        """
        old_cfg = Config.get_cfg("ToolDelta基本配置.json", {})
        old_cfg_keys = old_cfg.keys()
        need_upgrade_cfg = False
        for k, v in constants.LAUNCH_CFG.items():
            if k not in old_cfg_keys:
                old_cfg[k] = v
                need_upgrade_cfg = True
        if need_upgrade_cfg:
            Config.default_cfg("ToolDelta基本配置.json", old_cfg, True)
        return need_upgrade_cfg

    @staticmethod
    def welcome() -> None:
        """欢迎提示"""
        Print.print_with_info(
            "§dToolDelta Panel Embed By SuperScript", Print.INFO_LOAD
        )
        Print.print_with_info(
            "§dToolDelta Wiki: https://tooldelta-wiki.tblstudio.cn/", Print.INFO_LOAD
        )
        Print.print_with_info(
            "§dToolDelta 项目地址：https://github.com/ToolDelta", Print.INFO_LOAD
        )
        Print.print_with_info(
            f"§dToolDelta v {'.'.join([str(i) for i in VERSION])}", Print.INFO_LOAD
        )
        Print.print_with_info("§dToolDelta Panel 已启动", Print.INFO_LOAD)

    @staticmethod
    def basic_operation():
        """初始化文件夹"""
        os.makedirs(f"插件文件/{constants.TOOLDELTA_CLASSIC_PLUGIN}", exist_ok=True)
        os.makedirs(f"插件文件/{constants.TOOLDELTA_INJECTED_PLUGIN}", exist_ok=True)
        os.makedirs("插件配置文件", exist_ok=True)
        os.makedirs("tooldelta/neo_libs", exist_ok=True)
        os.makedirs("插件数据文件/game_texts", exist_ok=True)

    def add_console_cmd_trigger(
        self,
        triggers: list[str],
        arg_hint: str | None,
        usage: str,
        func: Callable[[list[str]], None],
    ) -> None:
        """注册 ToolDelta 控制台的菜单项

        Args:
            triggers (list[str]): 触发词列表
            arg_hint (str | None): 菜单命令参数提示句
            usage (str): 命令说明
            func (Callable[[list[str]], None]): 菜单回调方法
        """
        try:
            if self.consoleMenu.index(triggers) != -1:
                Print.print_war(f"§6后台指令关键词冲突: {func}, 不予添加至指令菜单")
        except Exception:
            self.consoleMenu.append([usage, arg_hint, func, triggers])

    def init_basic_help_menu(self, _) -> None:
        """初始化基本的帮助菜单"""
        menu = self.get_console_menus()
        Print.print_inf("§a以下是可选的菜单指令项: ")
        for usage, arg_hint, _, triggers in menu:
            if arg_hint:
                Print.print_inf(
                    f" §e{' 或 '.join(triggers)} §b{arg_hint} §f->  {usage}"
                )
            else:
                Print.print_inf(f" §e{' 或 '.join(triggers)}  §f->  {usage}")

    def comsole_cmd_start(self) -> None:
        """启动控制台命令"""
        Print.print_suc("ToolHack Terminal 进程已注入, 允许开启标准输入")

        def _try_execute_console_cmd(func, rsp, mode, arg1) -> int | None:
            try:
                if mode == 0:
                    rsp_arg = rsp.split()[1:]
                elif mode == 1:
                    rsp_arg = rsp[len(arg1) :].split()
            except IndexError:
                Print.print_err("[控制台执行命令] 指令缺少参数")
                return None
            try:
                return func(rsp_arg) or 0
            except Exception:
                Print.print_err(f"控制台指令出错: {traceback.format_exc()}")
                return 0

        @Utils.thread_func("控制台执行指令并获取回调")
        def _execute_mc_command_and_get_callback(cmd: str) -> None:
            """执行 Minecraft 指令并获取回调结果。

            Args:
                cmd (str): 要执行的 Minecraft 指令。

            Raises:
                ValueError: 当指令执行失败时抛出。
            """
            cmd = " ".join(cmd)
            try:
                result = self.link_game_ctrl.sendcmd_with_resp(cmd, 10)
                if (result.OutputMessages[0].Message == "commands.generic.syntax") | (
                    result.OutputMessages[0].Message == "commands.generic.unknown"
                ):
                    Print.print_err(f'未知的 MC 指令, 可能是指令格式有误: "{cmd}"')
                else:
                    mjon = self.link_game_ctrl.Game_Data_Handle.Handle_Text_Class1(
                        result.as_dict["OutputMessages"]
                    )
                    desc = json.dumps(
                        result.as_dict["OutputMessages"], indent=2, ensure_ascii=False
                    )
                    if result.SuccessCount:
                        print_str = "指令执行成功: " + " ".join(mjon)
                        Print.print_suc(print_str)
                        Print.print_suc(desc)
                    else:
                        print_str = "指令执行失败: " + " ".join(mjon)
                        Print.print_war(print_str)
                        Print.print_war(desc)

            except IndexError as exec_err:
                if isinstance(result, type(None)):
                    raise ValueError("指令执行失败") from exec_err
                if result.SuccessCount:
                    Print.print_suc(
                        f"指令执行成功, 详细返回结果:\n{json.dumps(result.as_dict['OutputMessages'], indent=2, ensure_ascii=False)}"
                    )
            except TimeoutError:
                Print.print_err("[超时] 指令获取结果返回超时")

        def _console_cmd_thread() -> None:
            """控制台线程"""
            self.add_console_cmd_trigger(
                ["?", "help", "帮助", "？"],
                None,
                "查询可用菜单指令",
                self.init_basic_help_menu,
            )
            self.add_console_cmd_trigger(
                ["exit"], None, "退出并关闭ToolDelta", lambda _: None
            )
            self.add_console_cmd_trigger(
                ["插件市场"],
                None,
                "进入插件市场",
                lambda _: plugin_market.market.enter_plugin_market(
                    self.plugin_market_url, in_game=True
                ),
            )
            self.add_console_cmd_trigger(
                ["/"], "[指令]", "执行 MC 指令", _execute_mc_command_and_get_callback
            )
            self.add_console_cmd_trigger(
                ["list"],
                None,
                "查询在线玩家",
                lambda _: Print.print_inf(
                    "在线玩家：" + ", ".join(self.link_game_ctrl.allplayers)
                ),
            )
            while 1:
                rsp = ""
                while True:
                    res = sys.stdin.read(1)
                    if res == "\n":  # 如果是换行符，则输出当前输入并清空输入
                        break
                    if res in ("", "^C", "^D"):
                        Print.print_inf("按退出键退出中...")
                        self.launcher.update_status(SysStatus.NORMAL_EXIT)
                        return
                    rsp += res
                for _, _, func, triggers in self.consoleMenu:
                    if not rsp:
                        continue
                    if rsp == "exit":
                        Print.print_inf("用户命令退出中...")
                        self.launcher.update_status(SysStatus.NORMAL_EXIT)
                        return
                    if rsp.split()[0] in triggers:
                        res = _try_execute_console_cmd(func, rsp, 0, None)
                        if res == -1:
                            return
                    else:
                        for tri in triggers:
                            if rsp.startswith(tri):
                                res = _try_execute_console_cmd(func, rsp, 1, tri)
                                if res == -1:
                                    return
                if res != 0 and rsp:
                    self.link_game_ctrl.say_to("@a", f"[§bToolDelta控制台§r] §3{rsp}§r")

        self.createThread(_console_cmd_thread, usage="控制台指令")

    def system_exit(self) -> None:
        """系统退出"""
        asyncio.run(safe_jump())
        self.link_plugin_group.execute_frame_exit(self.on_plugin_err)
        if not isinstance(self.launcher, (FrameNeOmgRemote,)):
            with contextlib.suppress(Exception):
                self.link_game_ctrl.sendwscmd(
                    f"/kick {self.link_game_ctrl.bot_name} ToolDelta 退出中。"
                )
            if not isinstance(self.launcher.neomg_proc, type(None)):
                self.launcher.neomg_proc.send_signal(signal.CTRL_BREAK_EVENT)
        if isinstance(self.launcher, (FrameNeOmgRemote, FrameNeOmg)):
            self.launcher.exit_event.set()

    def get_console_menus(self) -> list:
        """获取控制台命令列表

        Returns:
            list: 命令列表
        """
        return self.consoleMenu

    def set_game_control(self, game_ctrl: "GameCtrl") -> None:
        """使用外源 GameControl

        Args:help
            game_ctrl (_type_): GameControl 对象
        """
        self.link_game_ctrl = game_ctrl

    def set_plugin_group(self, plug_grp) -> None:
        """使用外源 PluginGroup

        Args:
            plug_grp (_type_): PluginGroup 对象
        """
        self.link_plugin_group: "PluginGroup" = plug_grp

    def get_game_control(self) -> "GameCtrl":
        """获取 GameControl 对象

        Returns:
            GameCtrl: GameControl 对象
        """
        gcl: "GameCtrl" = self.link_game_ctrl
        return gcl

    @staticmethod
    def safelyExit() -> None:
        """安全退出"""
        safe_close()
        publicLogger.exit()
        Print.print_inf("已保存数据与日志等信息。")


class GameCtrl:
    """游戏连接和交互部分"""

    def __init__(self, frame: "ToolDelta"):
        """初始化

        Args:
            frame (Frame): 继承 Frame 的对象
        """
        frame.basic_operation()
        self.Game_Data = GameTextsLoader().game_texts_data
        self.Game_Data_Handle = GameTextsHandle(self.Game_Data)
        self.linked_frame = frame
        self.players_uuid = {}
        self.allplayers = []
        self.all_players_data = {}
        self.linked_frame: ToolDelta
        self.pkt_unique_id: int = 0
        self.pkt_cache: list = []
        self.require_listen_packets = {9, 79, 63}
        self.store_uuid_pkt: dict[str, str] | None = None
        self.launcher = self.linked_frame.launcher
        if isinstance(self.launcher, (FrameNeOmgRemote, FrameNeOmg)):
            self.launcher.packet_handler = lambda pckType, pck: Utils.createThread(
                self.packet_handler, (pckType, pck), usage="数据包处理"
            )
        # 初始化基本函数
        self.sendcmd = self.launcher.sendcmd
        self.sendwscmd = self.launcher.sendwscmd
        self.sendwocmd = self.launcher.sendwocmd
        self.sendPacket = self.launcher.sendPacket
        if isinstance(self.linked_frame.launcher, FrameNeOmg):
            self.requireUUIDPacket = False
        else:
            self.requireUUIDPacket = True

    def set_listen_packets(self) -> None:
        """
        向启动器初始化监听的游戏数据包
        仅限内部调用
        """
        for pktID in self.require_listen_packets:
            self.launcher.add_listen_packets(pktID)

    def add_listen_pkt(self, pkt: int) -> None:
        """
        添加监听的数据包
        仅限内部调用

        Args:
            pkt (int): 数据包 ID
        """
        self.require_listen_packets.add(pkt)

    @Utils.thread_func("数据包处理方法")
    def packet_handler(self, pkt_type: int, pkt: dict) -> None:
        """数据包处理分发任务函数

        Args:
            pkt_type (int): 数据包类型
            pkt (dict): 数据包内容
        """
        is_skiped = self.linked_frame.link_plugin_group.processPacketFunc(pkt_type, pkt)
        if is_skiped:
            return
        if pkt_type == PacketIDS.PlayerList:
            self.process_player_list(pkt, self.linked_frame.link_plugin_group)
        elif pkt_type == PacketIDS.Text:
            self.process_text_packet(pkt, self.linked_frame.link_plugin_group)

    def process_player_list(self, pkt: dict, plugin_group: "PluginGroup") -> None:
        """处理玩家列表等数据包

        Args:
            pkt (dict): 数据包内容
            plugin_group (PluginGroup): 插件组对象
        """
        # 处理玩家进出事件
        res = self.launcher.get_players_and_uuids()
        if res:
            self.allplayers = list(res.keys())
            self.players_uuid.update(res)
        for player in pkt["Entries"]:
            isJoining = bool(player["Skin"]["SkinData"])
            playername = player["Username"]
            if isJoining and "§" in playername:
                self.say_to(
                    "@a",
                    "§l§7<§6§o!§r§l§7> §6此玩家名字中含特殊字符, 可能导致插件运行异常！",
                )
                self.all_players_data = self.launcher.omega.get_all_online_players()
                # 没有 VIP 名字供测试...
            if isJoining:
                self.tmp_tp_player(playername)
                Print.print_inf(f"§e{playername} 加入了游戏")
                self.all_players_data = self.launcher.omega.get_all_online_players()
                if playername not in self.allplayers and not res:
                    self.allplayers.append(playername)
                    return
                plugin_group.execute_player_join(
                    playername, self.linked_frame.on_plugin_err
                )
            else:
                playername = next(
                    (k for k, v in self.players_uuid.items() if v == player["UUID"]),
                    None,
                )
                if playername is None:
                    Print.print_war("无法获取 PlayerList 中玩家名字")
                    continue
                if playername != "???" and not res:
                    self.allplayers.remove(playername)
                Print.print_inf(f"§e{playername} 退出了游戏")
                self.all_players_data = self.launcher.omega.get_all_online_players()
                plugin_group.execute_player_leave(
                    playername, self.linked_frame.on_plugin_err
                )

    def process_text_packet(self, pkt: dict, plugin_grp: "PluginGroup") -> None:
        """处理 9 号数据包的消息

        Args:
            pkt (dict): 数据包内容
            plugin_grp (PluginGroup): 插件组对象
        """
        match pkt["TextType"]:
            case 2:
                if pkt["Message"] == "§e%multiplayer.player.joined":
                    player = pkt["Parameters"][0]
                    plugin_grp.execute_player_prejoin(
                        player, self.linked_frame.on_plugin_err
                    )
                elif not pkt["Message"].startswith(
                    "§e%multiplayer.player.joined"
                ) and not pkt["Message"].startswith("§e%multiplayer.player.left"):
                    jon = self.Game_Data_Handle.Handle_Text_Class1(pkt)
                    Print.print_inf(("§1" + " ".join(jon)))
                    if pkt["Message"].startswith("death."):
                        if len(pkt["Parameters"]) >= 2:
                            killer = pkt["Parameters"][1]
                        else:
                            killer = None
                        plugin_grp.execute_player_death(
                            pkt["Parameters"][0],
                            killer,
                            pkt["Message"],
                            self.linked_frame.on_plugin_err,
                        )
            case 1 | 7:
                player, msg = pkt["SourceName"], pkt["Message"]
                plugin_grp.execute_player_message(
                    player, msg, self.linked_frame.on_plugin_err
                )
                Print.print_inf(f"<{player}> {msg}")
            case 8:
                player, msg = pkt["SourceName"], pkt["Message"]
                Print.print_inf(f"{player} 使用 say 说：{msg.strip(f'[{player}]')}")
                plugin_grp.execute_command(player, msg, self.linked_frame.on_plugin_err)
            case 9:
                msg = pkt["Message"]
                try:
                    msg_text = json.loads(msg)["rawtext"]
                    if len(msg_text) > 0 and msg_text[0].get("translate") == "***":
                        Print.print_with_info("(该 tellraw 内容为敏感词)", "§f 消息 ")
                        return
                    msg_text = "".join([i["text"] for i in msg_text])
                    Print.print_with_info(msg_text, "§f 消息 ")
                except Exception:
                    pass

    def Inject(self) -> None:
        """载入游戏时的初始化"""
        if hasattr(self.launcher, "bot_name"):
            self.tmp_tp_all_players()
        res = self.launcher.get_players_and_uuids()
        self.all_players_data = self.launcher.omega.get_all_online_players()
        Utils.createThread(
            func=self.give_bot_effect_invisibility, usage="GiveBotEffectInvisibility"
        )
        if res:
            self.allplayers = list(res.keys())
            self.players_uuid.update(res)
        else:
            while 1:
                try:
                    cmd_result = self.sendcmd_with_resp("/list")
                    if (
                        cmd_result.OutputMessages is None
                        or len(cmd_result.OutputMessages) < 2
                    ):
                        raise ValueError
                    if (
                        cmd_result.OutputMessages[1].Parameters is None
                        or len(cmd_result.OutputMessages[1].Parameters) < 1
                    ):
                        raise ValueError
                    self.allplayers = (
                        cmd_result.OutputMessages[1].Parameters[0].split(", ")
                    )
                    break
                except (TimeoutError, ValueError):
                    Print.print_war("获取全局玩家失败..重试")
        self.linked_frame.comsole_cmd_start()
        self.linked_frame.link_plugin_group.execute_init(
            self.linked_frame.on_plugin_err
        )
        self.inject_welcome()

    def inject_welcome(self) -> None:
        """初始化欢迎信息"""
        Print.print_suc(
            "成功连接到游戏网络并初始化, 在线玩家: "
            + ", ".join(self.allplayers)
            + ", 机器人 ID: "
            + self.bot_name
        )
        self.sendcmd("/tag @s add robot")
        Print.print_inf(
            "在控制台输入 §b插件市场§r 以§a一键获取§rToolDelta官方和第三方的插件"
        )
        Print.print_suc("在控制台输入 §fhelp / ?§r§a 可查看控制台命令")

    @property
    def bot_name(self) -> str:
        if hasattr(self.launcher, "bot_name"):
            return self.launcher.get_bot_name()
        raise ValueError("此启动器框架无法产生机器人名")

    def sendcmd_with_resp(
        self, cmd: str, timeout: int | float = 30
    ) -> Packet_CommandOutput:
        resp: Packet_CommandOutput = self.sendwscmd(cmd, True, timeout)  # type: ignore
        return resp

    def say_to(self, target: str, text: str) -> None:
        """向玩家发送消息

        Args:
            target (str): 玩家名/目标选择器
            msg (str): 消息
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(f"tellraw {target} {text_json}")

    def player_title(self, target: str, text: str) -> None:
        """向玩家展示标题文本

        Args:
            target (str): 玩家名/目标选择器
            text (str): 文本
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(f"titleraw {target} title {text_json}")

    def player_subtitle(self, target: str, text: str) -> None:
        """向玩家展示副标题文本

        Args:
            target (str): 玩家名/目标选择器
            text (str): 文本
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(f"titleraw {target} subtitle {text_json}")

    def player_actionbar(self, target: str, text: str) -> None:
        """向玩家展示动作栏文本

        Args:
            target (str): 玩家名/目标选择器
            text (str): 文本
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(f"titleraw {target} actionbar {text_json}")

    def get_game_data(self) -> dict:
        """获取游戏常见字符串数据

        Returns:
            dict: 游戏常见字符串数据
        """
        return self.Game_Data

    def give_bot_effect_invisibility(self) -> None:
        """给机器人添加隐身效果"""
        start_time = time.time()
        while self.linked_frame.link_game_ctrl.launcher.status == SysStatus.RUNNING:
            if time.time() - start_time > 16384:
                self.sendwocmd(f"effect {self.bot_name} invisibility 99999 255 true")
                start_time = time.time()
            time.sleep(1)

    def tmp_tp_all_players(self) -> None:
        """
        内部调用：在 ToolDelta 连接到 NeOmega 后先 tp 所有玩家（防止玩家 entityruntimeid 为空）
        外部调用：临时传送至所有玩家后回到原位
        """
        BotPos: tuple[float, float, float] = getPosXYZ(self.bot_name)
        for player in self.allplayers:
            if player == self.bot_name:
                continue
            self.sendwocmd(f"tp {self.bot_name} {player}")
        self.sendwocmd(
            f"tp {self.bot_name} {str(int(BotPos[0])) + ' ' + str(int(BotPos[1])) + ' ' + str(int(BotPos[2]))}"
        )

    def tmp_tp_player(self, player: str) -> None:
        """临时传送玩家 (看一眼玩家就回来，多数用于捕获数据)
        Args:

            player (str): 玩家名

        """
        BotPos: tuple[float, float, float] = getPosXYZ(self.bot_name)
        self.sendwocmd(f"tp {self.bot_name} {player}")
        self.sendwocmd(
            f"tp {self.bot_name} {str(int(BotPos[0])) + ' ' + str(int(BotPos[1])) + ' ' + str(int(BotPos[2]))}"
        )
